from typing import Iterable, Union

from airium import Airium
from natsort import natsorted
from yarl import URL

from .util import canonicalize_name
from .wheel_file import WheelFile


# modified from https://github.com/repo-helper/simple503/blob/master/simple503/__init__.py
def generate_index(projects: Iterable[str]) -> Airium:
    """
    Generate the simple repository index page, containing a list of all projects.

    :param projects: The list of projects to generate links for.
    """

    base_url = URL("./")
    index = Airium()

    index("<!DOCTYPE html>")
    with index.html(lang="en"):
        with index.head():
            get_meta_tags(index)

            with index.title():
                index("Simple Package Repository")

        with index.body():
            for project_name in natsorted(projects, key=str.lower):
                normalized_name = canonicalize_name(project_name)

                with index.a(href=f"{base_url / normalized_name}/"):
                    index(project_name)
                index.br()

    return index


def generate_project_page(
    name: str, files: Iterable[WheelFile], base_url: Union[str, URL] = "/"
) -> Airium:
    """
    Generate the repository page for a project.

    :param name: The project name, e.g. ``domdf-python-tools``.
    :param files: An iterable of files for the project, which will be linked to from the index page.
    :param base_url: The base URL of the Python package repository.
        For example, with PyPI's URL, a URL of /foo/ would be https://pypi.org/simple/foo/.
    """

    name = canonicalize_name(name)
    page = Airium()

    page("<!DOCTYPE html>")
    with page.html(lang="en"):
        with page.head():
            get_meta_tags(page)
            with page.title():
                page(f"Links for {name}")

        with page.body():
            with page.h1():
                # Not part of the spec, but allowed
                page(f"Links for {name}")

            for wheel_file in files:
                wheel_file.as_anchor(page, base_url)
                page.br()

    return page


def get_meta_tags(page: Airium) -> None:
    from . import __version__

    # Not part of the spec, but allowed
    page.meta(name="generator", content=f"index503 version {__version__}")
    page.meta(name="pypi:repository-version", content="1.0")
    page.meta(charset="UTF-8")



import logging
import os
import tempfile
from pathlib import Path

_LOGGER = logging.getLogger(__name__)


# from https://github.com/home-assistant/core/blob/dev/homeassistant/util/file.py
def write_utf8_file(
    filename: Path,
    utf8_data: str,
    private: bool = False,
) -> None:
    """Write a file and rename it into place.

    Writes all or nothing.
    """

    tmp_filename = ""
    try:
        # Modern versions of Python tempfile create this file with mode 0o600
        with tempfile.NamedTemporaryFile(
            mode="w", encoding="utf-8", dir=os.path.dirname(filename), delete=False
        ) as fdesc:
            fdesc.write(utf8_data)
            tmp_filename = fdesc.name
            if not private:
                os.fchmod(fdesc.fileno(), 0o644)
        os.replace(tmp_filename, filename)
    finally:
        if os.path.exists(tmp_filename):
            try:
                os.remove(tmp_filename)
            except OSError as err:
                # If we are cleaning up then something else went wrong, so
                # we should suppress likely follow-on errors in the cleanup
                _LOGGER.error(
                    "File replacement cleanup failed for %s while saving %s: %s",
                    tmp_filename,
                    filename,
                    err,
                )



import logging
from pathlib import Path
from typing import Optional

from dist_meta import distributions

_LOGGER = logging.getLogger(__name__)


def extract_metadata_from_wheel_file(wheel_path: Path) -> Optional[str]:
    """Extract the METADATA file from a wheel file."""
    with distributions.WheelDistribution.from_path(wheel_path) as wd:
        if not wd.has_file("METADATA"):  # pragma: no cover
            _LOGGER.warning(f"METADATA file not found in {wheel_path}")
            return None
        return wd.read_file("METADATA")



from pathlib import Path

import click
from consolekit import click_command

from .index import make_index


@click.argument("origin", type=click.STRING)
@click_command()
def main_cli(origin: str) -> None:
    origin_path = Path(origin)
    if not origin_path.exists():
        raise FileNotFoundError(f"Directory {origin_path} does not exist")
    target_path = make_index(origin_path)
    print(f"Index generated at {target_path}")



# https://github.com/repo-helper/simple503/blob/master/simple503/__init__.py
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
import posixpath
from dataclasses import asdict, dataclass
from html import escape
from pathlib import Path
from typing import Any, Dict, Optional, Union

from airium import Airium
from dist_meta import metadata
from typing_extensions import Literal
from yarl import URL

from .metadata import extract_metadata_from_wheel_file
from .util import canonicalize_name, get_sha256_hash

WHEEL_FILE_VERSION = 9

HASH_FORMAT = "sha256"


@dataclass
class WheelFile:
    """Represents a wheel file in the repository."""

    version: int
    """The version of the wheel file format."""

    canonical_name: str
    """The canonicalized name of the project."""

    filename: str
    """The filename of the wheel file."""

    metadata_name: str
    """The Name field from the METADATA file."""

    wheel_hash: str  # HASH_FORMAT
    """
    The hash of the wheel file.

    Repositories SHOULD choose a hash function from one of the ones guaranteed
    to be available via the hashlib module in the Python standard library
    (currently ``md5``, ``sha1``, ``sha224``, ``sha256``, ``sha384``, ``sha512``).
    The current recommendation is to use ``sha256``.
    """

    requires_python: Optional[str] = None
    """
    The ``Requires-Python`` attribute from the wheel's ``METADATA`` file.

    :py:obj:`None` if undefined.
    """

    metadata_hash: Union[str, Literal[True], None] = None
    """
    The hash of the wheel's METADATA file.

    :py:obj:`None` if the metadata file is not exposed.
    May be :py:obj:`True` if no hash is available.
    """

    def as_anchor(self, page: Airium, base_url: Union[str, URL] = "/") -> None:
        """
        Generate an anchor tag in a :class:`airium.Airium` document for this file.

        :param page:
        :param base_url: The base URL of the Python package repository.
        """

        base_url = URL(base_url)

        href = f"{base_url / self.filename}#{HASH_FORMAT}={self.wheel_hash}"
        kwargs = {"href": href}

        if self.requires_python is not None:
            kwargs["data-requires-python"] = escape(self.requires_python)
        elif self.metadata_hash is not None:
            kwargs["data-dist-info-metadata"] = f"{HASH_FORMAT}={self.metadata_hash}"

        with page.a(**kwargs):
            page(posixpath.basename(self.filename))

    def update_metadata(self, metadata_path: Path) -> None:
        """
        Update the metadata hash of this wheel file.

        :param metadata_path:
        """

        self.metadata_hash = get_sha256_hash(metadata_path)

    def as_dict(self) -> Dict[str, Any]:
        """Return a dictionary representation of this wheel file."""
        return asdict(self)

    @classmethod
    def from_cache(cls, cache_data: Dict[str, Any]) -> Optional["WheelFile"]:
        """Create a :class:`~.WheelFile` from a cache entry."""
        if cache_data["version"] != WHEEL_FILE_VERSION:
            return None
        return cls(**cache_data)

    @classmethod
    def from_wheel(
        cls,
        wheel_path: Path,
        metadata_path: Path,
    ) -> Optional["WheelFile"]:
        """Create a :class:`~.WheelFile` from a wheel file and its metadata file."""
        metadata_string = extract_metadata_from_wheel_file(wheel_path)
        if not metadata_string:
            return None
        metadata_path.write_text(metadata_string)
        wheel_metadata = metadata.loads(metadata_string)
        wheel_file_name = wheel_path.name
        metadata_name = wheel_metadata["Name"]
        canonical_name = canonicalize_name(metadata_name)
        return cls(
            version=WHEEL_FILE_VERSION,
            metadata_name=metadata_name,
            canonical_name=canonical_name,
            filename=wheel_file_name,
            wheel_hash=get_sha256_hash(wheel_path),
            requires_python=wheel_metadata.get("Requires-Python"),
            metadata_hash=get_sha256_hash(metadata_path),
        )



import fcntl
import json
import logging
import re
from contextlib import contextmanager
from hashlib import sha256
from pathlib import Path
from typing import Any, Dict, Generator

_LOGGER = logging.getLogger(__name__)

_canonicalize_regex = re.compile(r"[-_.]+")


# PEP 427: The build number must start with a digit.
def canonicalize_name(name: str) -> str:
    """Canonicalize a name."""
    # This is taken from PEP 503.
    return _canonicalize_regex.sub("-", name).lower()


def get_sha256_hash(filename: Path) -> str:
    """Get SHA256 hash of a file."""
    with filename.open("rb") as f:
        bytes = f.read()  # read entire file as bytes
        return sha256(bytes).hexdigest()


def load_json_file(filename: Path) -> Dict[str, Dict[str, Any]]:
    """Get a json file."""
    with filename.open("rb") as f:
        bytes = f.read()  # read entire file as bytes
        return json.loads(bytes)


@contextmanager
def exclusive_lock(origin_path: Path) -> Generator[None, None, None]:
    parent = origin_path.parent
    lock_file = parent / (f".{origin_path.name}.index_503.lock")
    with lock_file.open("a") as fh:
        try:
            fcntl.flock(fh, fcntl.LOCK_EX | fcntl.LOCK_NB)
            yield
            return
        except OSError:
            # Another instance is running, wait for it to finish
            _LOGGER.warning("Another instance is running, waiting")
        fcntl.flock(fh, fcntl.LOCK_EX)
        _LOGGER.warning("Another instance finished, continuing")
        yield



import json
import logging
from pathlib import Path
from typing import Any, Dict, Set

from .file import write_utf8_file
from .util import load_json_file

_LOGGER = logging.getLogger(__name__)

CACHE_FILE = "cache.json"


class IndexCache:
    def __init__(self, target_path: Path) -> None:
        """Cache of WheelFiles between runs."""
        cache_file = target_path.joinpath(CACHE_FILE)
        self.cache_file = cache_file
        self.cache: Dict[str, Dict[str, Any]] = {}

    def load(self) -> None:
        """Load the cache from a file."""
        if self.cache_file.exists():
            self.cache = load_json_file(self.cache_file)

    def write_to_new(self, target: Path) -> None:
        """Write the cache to a new file."""
        new_cache_file = target.joinpath(CACHE_FILE)
        write_utf8_file(new_cache_file, json.dumps(self.cache))

    def remove_stale_keys(self, all_wheel_files: Set[str]) -> None:
        """Remove any wheel file names that no longer exist."""
        removed_wheels = set(self.cache.keys()) - all_wheel_files
        for old_wheel_file in removed_wheels:
            del self.cache[old_wheel_file]



import glob
import logging
import os
from collections import defaultdict
from operator import attrgetter
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from typing import Dict, List, Set

from natsort import natsorted
from yarl import URL

from .cache import IndexCache
from .file import write_utf8_file
from .page_generator import generate_index, generate_project_page
from .util import canonicalize_name, exclusive_lock
from .wheel_file import WheelFile

_LOGGER = logging.getLogger(__name__)


def make_index(origin_path: Path) -> Path:
    """Generate a simple repository of Python wheels.

    This function will take a directory of wheels at the top level
    of a webserver and generate a simple repository of wheels.

    :param origin: The name of the directory containing the wheels.

    Example
    musllinux

    This will generate
    musllinux-index
    """
    with exclusive_lock(origin_path):
        return IndexMaker(origin_path).make_index()


class IndexMaker:
    """Generate a simple repository of Python wheels."""

    def __init__(self, origin_path: Path) -> None:
        """Generate a simple repository of Python wheels."""
        self.origin_path = origin_path
        self.origin_name = origin_path.name
        target_path = origin_path.parent / (origin_path.name + "-index")
        self.target_path = target_path
        self.cache = IndexCache(target_path)

    def make_index(self) -> Path:
        """Generate a simple repository of Python wheels."""
        target_path = self.target_path
        old_index = target_path.readlink() if target_path.exists() else None
        temp_dir = mkdtemp(None, None, str(self.target_path.parent))
        try:
            self.cache.load()
            temp_dir_path = Path(temp_dir)
            temp_dir_path.chmod(0o755)

            self._make_index_at_temp_dir(temp_dir_path)
            self._atomic_replace_old_index(temp_dir_path, target_path)

            if old_index:
                _LOGGER.debug("Removing old index %s", old_index)
                rmtree(old_index)

            return target_path
        except Exception:
            _LOGGER.exception("Error generating index")
            rmtree(temp_dir)
            raise

    def _atomic_replace_old_index(self, temp_dir_path: Path, target_path: Path) -> None:
        """Atomically replace the old index with the new one."""
        final_name = target_path.parent / (target_path.name + "-" + temp_dir_path.name)
        final_build_name = final_name.parent / (final_name.name + "-build")

        _LOGGER.debug("Renaming %s with %s", temp_dir_path, final_name)
        # Rename the new index to the final name
        os.rename(temp_dir_path, final_name)

        _LOGGER.debug("Symlinking %s with %s", final_name, final_build_name)
        # Create a temporary symlink to the final name
        os.symlink(final_name, final_build_name)

        _LOGGER.debug("Replacing %s with %s", final_build_name, target_path)
        # Finally replace the live index with the new one
        os.replace(final_build_name, target_path)

    def _make_index_at_temp_dir(self, temp_dir_path: Path) -> None:
        """Generate a simple repository of Python wheels in a temp dir."""
        new_wheel_file_objects: List[WheelFile] = []
        projects: Dict[str, List[WheelFile]] = defaultdict(list)
        wheel_file_name_to_metadata_path: Dict[str, Path] = {}
        all_wheel_files: Set[str] = set()
        canonical_name_to_metadata_name: Dict[str, str] = {}
        raw_cache = self.cache.cache

        for wheel_file in glob.glob(str(self.origin_path.joinpath("*.whl"))):
            wheel_path = Path(wheel_file)
            wheel_file_name = wheel_path.name
            all_wheel_files.add(wheel_file_name)
            target_file = temp_dir_path.joinpath(wheel_file_name)
            metadata_path = target_file.with_suffix(f"{target_file.suffix}.metadata")
            wheel_file_symlink_target = f"../{self.origin_name}/{wheel_file_name}"

            if (wheel_cache := raw_cache.get(wheel_file_name)) and (
                wheel_file_obj := WheelFile.from_cache(wheel_cache)
            ):
                os.link(self.target_path.joinpath(metadata_path.name), metadata_path)
            elif wheel_file_obj := WheelFile.from_wheel(wheel_path, metadata_path):
                wheel_file_name_to_metadata_path[wheel_file_name] = metadata_path
                new_wheel_file_objects.append(wheel_file_obj)
                raw_cache[wheel_file_name] = wheel_file_obj.as_dict()
            else:
                continue

            canonical_name = wheel_file_obj.canonical_name
            metadata_name = wheel_file_obj.metadata_name
            projects[metadata_name].append(wheel_file_obj)
            canonical_name_to_metadata_name[canonical_name] = metadata_name
            os.symlink(wheel_file_symlink_target, target_file)

        self.cache.remove_stale_keys(all_wheel_files)
        self.generate_index_pages(temp_dir_path, projects)
        self.cache.write_to_new(temp_dir_path)

    def generate_index_pages(
        self, temp_dir_path: Path, projects: Dict[str, List[WheelFile]]
    ) -> None:
        """Generate the index pages."""
        index_content = str(generate_index(projects.keys()))
        write_utf8_file(temp_dir_path.joinpath("index.html"), index_content)
        project_base_url = URL("../")

        for project_name, project_files in projects.items():
            project_dir: Path = temp_dir_path.joinpath(canonicalize_name(project_name))
            project_dir.mkdir(exist_ok=True, mode=0o755)
            project_index = generate_project_page(
                project_name,
                natsorted(project_files, key=attrgetter("filename"), reverse=True),
                project_base_url,
            )

            write_utf8_file(project_dir.joinpath("index.html"), str(project_index))

        _LOGGER.debug("Generated index pages for %s projects.", len(projects))
